/*
 * Copyright (C) 2013 Philippe Komma
 * 
 * This file is part of Number: A Number Detection and 
 * Classification Framework
 * 
 * Number is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Foobar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Foobar. If not, see <http://www.gnu.org/licenses/>.
 */

#include "Classifier.h"

#include <stdio.h>
#include <string.h>

#include <fstream>
using std::ifstream;
#include <iostream>
using std::cout;
using std::endl;


Classifier::Classifier() :
	m_number( NULL ),
	m_scaled( NULL ),
	m_bitCount( NULL )
{
}


Classifier::Classifier( const char* fname )
{
	init( fname );
}


Classifier::~Classifier()
{
	release();
}


bool Classifier::init( const char* fname )
{
	m_scaled = new unsigned char[ 32ul * 4 ];

	m_bitCount = new unsigned char[ 256 ];

	int i, b;
	for ( i = 0; i <= 255; i++ ) {
		m_bitCount[ i ] = 0;
		for ( b = 0; b < 8; b++ ) {
			m_bitCount[ i ] += ( i >> b ) & 1;
		}
	}

	return load( fname );
}


bool Classifier::load( const char* fname )
{
	ifstream fs( fname );

	if ( !fs.is_open() ) {
		cout << "error loading " << fname << endl;
		return false;
	}

	m_number = new Number*[ 10 * 7 ];

	int n, i;
	for ( n = 0; n < 70; n++ ) {
		m_number[ n ] = new Number;
		m_number[ n ]->data = new unsigned char[ 128 ];
		fs >> m_number[ n ]->value;
		for ( i = 0; i < 128; i++ ) {
			int bits;
			fs >> bits;
			m_number[ n ]->data[ i ] = bits;
		}
	}

	fs.close();

	return true;
}


void Classifier::release()
{
	if ( m_scaled ) {
		delete [] m_scaled;
		m_scaled = NULL;
	}

	if ( m_bitCount ) {
		delete [] m_bitCount;
		m_bitCount = NULL;
	}

	if ( m_number ) {
		int i;
		for ( i = 0; i < 70; i++ ) {
			delete [] m_number[ i ]->data;
			delete m_number[ i ];
		}
		delete [] m_number;
		m_number = NULL;
	}
}


int Classifier::linspace( int a, int b, int n, int idx )
{
	return a + ( b - a ) * ( idx - 1 ) / ( n - 1 );
}


void Classifier::scalePattern( unsigned int label, unsigned int* labelBuffer, int widthStep, int x1, int y1, int x2, int y2 )
{
	memset( m_scaled, 0, 32ul * 4 );

	int x, y;  
	int dstOffset = 0;  
	unsigned char v = 1;
	for ( y = 0; y < 32; y++ ) {
		for ( x = 0; x < 32; x++ ) {
			int ix = linspace( x1, x2, 32, x + 1 );
			int iy = linspace( y1, y2, 32, y + 1 );
			int srcOffset = ix + iy * widthStep;

			if ( label == labelBuffer[ srcOffset ] ) {
				m_scaled[ dstOffset ] |= v;
			} 

			v <<= 1;
			if ( !v ) {
				v = 1;
				dstOffset++;
			}
		}
	}
}


void Classifier::classify( unsigned int label, unsigned int* labelBuffer, int widthStep, int x1, int y1, int x2, int y2, int& number, int& dist )
{  
	scalePattern( label, labelBuffer, widthStep, x1, y1, x2, y2 );

	unsigned int i;
	unsigned int minDst = 0xffffffff;  
	unsigned int minIdx = -1;

	for ( i = 0; i < 70; i++ ) {
		unsigned int dist = calcDist( m_scaled, m_number[ i ]->data );
		if ( dist < minDst ) {
			minDst = dist;
			minIdx = i;
		}
	}

	dist = minDst;

	if ( minDst > 180 ) {
		number = -1;
	} else {
		number = m_number[ minIdx ]->value;
	}
}


unsigned int Classifier::calcDist( unsigned char* ref, unsigned char* cmp )
{
	int i;
	unsigned int dist = 0;
	for ( i = 0; i < 128; i++ ) {
		unsigned char difference = ref[ i ] ^ cmp[ i ];
		dist += m_bitCount[ difference ];
	}

	return dist;
}